package net.knarcraft.blacksmith.trait;

import net.citizensnpcs.api.npc.NPC;
import net.knarcraft.blacksmith.config.SalvageMethod;
import net.knarcraft.blacksmith.config.scrapper.ScrapperNPCSettings;
import net.knarcraft.blacksmith.manager.EconomyManager;
import net.knarcraft.blacksmith.util.ItemHelper;
import net.knarcraft.blacksmith.util.SalvageHelper;
import org.bukkit.entity.Damageable;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.jetbrains.annotations.NotNull;

import java.util.Calendar;
import java.util.List;
import java.util.Objects;
import java.util.Random;

import static net.knarcraft.blacksmith.formatting.BlacksmithStringFormatter.sendNPCMessage;
import static net.knarcraft.blacksmith.util.ItemHelper.updateDamage;

/**
 * A representation of the session between a player and a scrapper
 */
public class SalvageSession extends Session implements Runnable {

    private final ScrapperTrait scrapperTrait;
    private final NPC npc;
    private final ItemStack itemToSalvage;
    private final ScrapperNPCSettings config;
    private final List<ItemStack> salvage;
    private final int enchantmentLevels;
    protected final SalvageMethod salvageMethod;
    private static final Random random = new Random();

    /**
     * Instantiates a new session
     *
     * @param scrapperTrait <p>A reference to the scrapper trait</p>
     * @param player        <p>The player initiating the session</p>
     * @param npc           <p>The scrapper NPC involved in the session</p>
     * @param config        <p>The config to use for the session</p>
     * @param salvageMethod <p>The salvage method performed in this session</p>
     */
    SalvageSession(@NotNull ScrapperTrait scrapperTrait, @NotNull Player player, @NotNull NPC npc,
                   @NotNull ScrapperNPCSettings config, @NotNull List<ItemStack> salvage,
                   @NotNull SalvageMethod salvageMethod) {
        super(player);
        this.scrapperTrait = scrapperTrait;
        this.npc = npc;
        this.itemToSalvage = player.getInventory().getItemInMainHand();
        this.config = config;
        this.salvage = salvage;
        this.enchantmentLevels = SalvageHelper.getTotalEnchantmentLevels(this.itemToSalvage);
        this.salvageMethod = salvageMethod;
    }

    @Override
    public boolean isSessionInvalid() {
        // Prevent player from switching items during session
        ItemStack itemInHand = this.player.getInventory().getItemInMainHand();
        if (!itemToSalvage.equals(itemInHand)) {
            sendNPCMessage(this.npc, this.player, this.config.getItemChangedMessage());
            return true;
        }

        if (EconomyManager.cannotPayForSalvage(this.player, this.salvageMethod)) {
            sendNPCMessage(this.npc, this.player, this.config.getInsufficientFundsMessage());
            return true;
        }
        return false;
    }

    @Override
    protected int getActionDelay() {
        if (this.config.getMaxSalvageDelay() > 0) {
            //Finish the salvaging after a random delay between the max and min
            return new Random().nextInt(this.config.getMaxSalvageDelay()) + this.config.getMinSalvageDelay();
        } else {
            //Finish the salvaging as soon as possible
            return 0;
        }
    }

    /**
     * Runs the actual salvage which fixes the item and gives it back to the player
     */
    @Override
    public void run() {
        sendNPCMessage(this.npc, this.player, salvageItem() ? this.config.getSuccessMessage() : this.config.getFailMessage());

        //Stop the reforged item from displaying in the scrapper's hand
        if (this.npc.getEntity() instanceof Player) {
            ((Player) this.npc.getEntity()).getInventory().setItemInMainHand(null);
        } else {
            Objects.requireNonNull(((LivingEntity) this.npc.getEntity()).getEquipment()).setItemInMainHand(null);
        }

        //Mark this scrapper as available
        this.scrapperTrait.unsetSession();

        // Start cool-down
        Calendar wait = Calendar.getInstance();
        wait.add(Calendar.SECOND, this.config.getSalvageCoolDown());
        this.scrapperTrait.addCoolDown(this.player.getUniqueId(), wait);
    }

    /**
     * Trues to salvage an item, and gives the return item
     *
     * @return <p>True if the salvage was successful. False otherwise.</p>
     */
    private boolean salvageItem() {
        if (random.nextInt(100) < this.config.getFailChance()) {
            failSalvage();
            return false;
        } else {
            // Give the player the result of the salvage
            giveSalvage();
            return true;
        }
    }

    /**
     * The method to run when a crapper fails salvaging an item
     */
    private void failSalvage() {
        if (this.itemToSalvage.getItemMeta() instanceof Damageable) {
            //Damage the item
            short currentItemDurability = ItemHelper.getDurability(this.itemToSalvage);
            short newDurability = (short) (currentItemDurability + (currentItemDurability * random.nextInt(8)));
            short maxDurability = this.itemToSalvage.getType().getMaxDurability();
            if (newDurability <= 0) {
                newDurability = (short) (maxDurability / 3);
            } else if (currentItemDurability + newDurability > maxDurability) {
                newDurability = (short) (maxDurability - random.nextInt(maxDurability - 25));
            }
            updateDamage(this.itemToSalvage, maxDurability - newDurability);
        } else {
            this.itemToSalvage.setAmount(Math.max(this.itemToSalvage.getAmount() - 1, 1));
        }
    }

    /**
     * Gives the player the calculated salvage
     */
    private void giveSalvage() {
        // TODO: Find a better calculation than 1 enchantment level = 1 exp level
        // Gives the player back some of the EXP used on an item
        this.player.giveExpLevels(this.enchantmentLevels);
        if (this.config.getDropItem() || !this.player.isOnline() || this.player.getInventory().firstEmpty() == -1) {
            // If the player isn't online, or the player cannot fit the item, drop the item to prevent it from
            // disappearing, even if drop item is disabled.
            for (ItemStack item : this.salvage) {
                this.player.getWorld().dropItemNaturally(this.npc.getEntity().getLocation(), item);
            }
        } else {
            for (ItemStack item : this.salvage) {
                if (ItemHelper.canFitItem(this.player.getInventory(), item)) {
                    this.player.getInventory().addItem(item);
                } else {
                    // If the player cannot fit all the salvage, drop it on the ground
                    this.player.getWorld().dropItemNaturally(this.player.getLocation(), item);
                }
            }
        }
    }

}
